@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2000-2006  Ulric Eriksson <ulric@siag.nu>
+   Copyright (C) 2000-2008  Ulric Eriksson <ulric@siag.nu>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -80,6 +80,7 @@ static SSL_CTX *ssl_context = NULL;
 #define BLACKLIST_TIME	30	/* how long to shun a server that is down */
 #define TRACKING_TIME	0	/* how long a client is remembered */
 #define KEEP_MAX	100	/* how much to keep from the URI */
+#define WEIGHT_FACTOR	256	/* to make weight kick in earlier */
 
 typedef struct {
 	int downfd, upfd;
@@ -102,6 +103,7 @@ typedef struct {
 	struct in_addr addr;
 	int c;			/* connections */
 	int weight;		/* default 1 */
+	int prio;
 	int maxc;		/* max connections, soft limit */
 	int hard;		/* max connections, hard limit */
 	unsigned long long sx, rx;	/* bytes sent, received */
@@ -148,6 +150,7 @@ static int blacklist_time = BLACKLIST_TIME;
 static int tracking_time = TRACKING_TIME;
 static int roundrobin = 0;
 static int weight = 0;
+static int prio = 0;
 static int hash = 0;
 static int stubborn = 0;
 static int nblock = 1;
@@ -499,6 +502,7 @@ static void webstats(void)
 		"<td bgcolor=\"#80f080\">sent</td>\n"
 		"<td bgcolor=\"#80f080\">received</td>\n"
 		"<td bgcolor=\"#80f080\">weight</td>\n"
+		"<td bgcolor=\"#80f080\">prio</td>\n"
 		"</tr>\n");
 	for (i = 0; i < nservers; i++) {
 		fprintf(fp,
@@ -513,12 +517,13 @@ static void webstats(void)
 			"<td>%llu</td>\n"
 			"<td>%llu</td>\n"
 			"<td>%d</td>\n"
+			"<td>%d</td>\n"
 			"</tr>\n",
 			i, inet_ntoa(servers[i].addr),
 			servers[i].status, servers[i].port,
 			servers[i].c, servers[i].maxc, servers[i].hard,
 			servers[i].sx, servers[i].rx,
-			servers[i].weight);
+			servers[i].weight, servers[i].prio);
 	}
 	fprintf(fp, "</table>\n");
 
@@ -767,22 +772,23 @@ static void setipaddress(struct in_addr *a, char *p)
 }
 
 static void setaddress(struct in_addr *a, int *port, char *s,
-		int dp, int *maxc, int *hard, int *weight)
+		int dp, int *maxc, int *hard, int *weight, int *prio)
 {
 	struct hostent *h;
 	char address[1024], pno[100];
-	int n = sscanf(s, "%999[^:]:%99[^:]:%d:%d:%d",
-		address, pno, maxc, hard, weight);
+	int n = sscanf(s, "%999[^:]:%99[^:]:%d:%d:%d:%d",
+		address, pno, maxc, hard, weight, prio);
 
 	if (n > 1) *port = getport(pno);
 	else *port = dp;
 	if (n < 3) *maxc = 0;
 	if (n < 4) *hard = 0;
 	if (n < 5) *weight = 0;
+	if (n < 6) *prio = 0;
 
 	if (debuglevel)
-		debug("n = %d, address = %s, pno = %d, maxc1 = %d, hard = %d, weight = %d",
-			n, address, *port, *maxc, *hard, *weight);
+		debug("n = %d, address = %s, pno = %d, maxc1 = %d, hard = %d, weight = %d, prio = %d",
+			n, address, *port, *maxc, *hard, *weight, *prio);
 
 	if (!(h = gethostbyname(address))) {
 		if ((a->s_addr = inet_addr(address)) == -1) {
@@ -828,7 +834,7 @@ static void netlog(int fd, int i, unsigned char *r, int n)
 
     client_ip timestamp server_ip request
 */
-static void log(FILE *fp, int i, unsigned char *b, int n)
+static void log_request(FILE *fp, int i, unsigned char *b, int n)
 {
 	int j;
 	if (n > KEEP_MAX) n = KEEP_MAX;
@@ -894,7 +900,7 @@ static int copy_up(int i)
 	int to = conns[i].upfd;
 	int serverindex = conns[i].index;
 
-	char b[BUFFER_MAX];
+	unsigned char b[BUFFER_MAX];
 
 #ifdef HAVE_SSL
 	SSL *ssl = conns[i].ssl;
@@ -928,8 +934,8 @@ static int copy_up(int i)
 		if (debuglevel > 2) dump(b, rc);
 
 		if (logfp) {
-			log(logfp, i, b, rc);
-			if (debuglevel > 2) log(stderr, i, b, rc);
+			log_request(logfp, i, b, rc);
+			if (debuglevel > 2) log_request(stderr, i, b, rc);
 		}
 		if (logsock != -1) {
 			netlog(logsock, i, b, rc);
@@ -967,7 +973,7 @@ static int copy_down(int i)
 	SSL *ssl = conns[i].ssl;
 #endif
 
-	char b[BUFFER_MAX];
+	unsigned char b[BUFFER_MAX];
 
 	rc = read(from, b, BUFFER_MAX);	/* no ssl here */
 
@@ -1117,7 +1123,7 @@ static void usage(void)
 	       "	  [-E certfile] [-K keyfile] \\\n"
 	       "	  [-G cacertfile] [-A cacertdir] \\\n"
 	       "	  [-Z] [-R] [-L protocol] \\\n"
-	       "	  [host:]port h1[:p1[:maxc1[:hard1[:weight1]]]] [h2[:p2[:maxc2[:hard2[:weight2]]]]] ...\n"
+	       "	  [host:]port h1[:p1[:maxc1[:hard1[:weight1[:prio1]]]]] [h2[:p2[:maxc2[:hard2[:weight2[:prio2]]]]]] ...\n"
 	       "\n"
 	       "  -B host:port abuse server for naughty clients\n"
 	       "  -C port   control port\n"
@@ -1174,7 +1180,11 @@ static void background(void)
 		if (childpid > 0) {
 			exit(0);	/* parent */
 		}
-	}
+	}	
+	int devnull_fd = open("/dev/null", O_RDWR);
+	dup2(devnull_fd,0); /* stdin */
+	dup2(devnull_fd,1); /* stdout */
+	dup2(devnull_fd,2); /* stderr */
 	setsid();
 	signal(SIGCHLD, SIG_IGN);
 #endif
@@ -1208,7 +1218,7 @@ static void init(int argc, char **argv)
 		setaddress(&servers[server].addr, &servers[server].port,
 			argv[i], port,
 			&servers[server].maxc, &servers[server].hard,
-			&servers[server].weight);
+			&servers[server].weight, &servers[server].prio);
 		servers[server].sx = 0;
 		servers[server].rx = 0;
 
@@ -1221,7 +1231,7 @@ static void init(int argc, char **argv)
 		setaddress(&servers[server].addr, &servers[server].port,
 			"0.0.0.0", 0,
 			&servers[server].maxc, &servers[server].hard,
-			&servers[server].weight);
+			&servers[server].weight, &servers[server].prio);
 		servers[server].sx = 0;
 		servers[server].rx = 0;
 
@@ -1235,7 +1245,7 @@ static void init(int argc, char **argv)
 		setaddress(&servers[server].addr, &servers[server].port,
 			   e_server, port,
 			   &servers[server].maxc, &servers[server].hard,
-			   &servers[server].weight);
+			   &servers[server].weight, &servers[server].prio);
 		servers[server].sx = 0;
 		servers[server].rx = 0;
 		server++;
@@ -1248,7 +1258,7 @@ static void init(int argc, char **argv)
 		setaddress(&servers[server].addr, &servers[server].port,
 			   a_server, port,
 			   &servers[server].maxc, &servers[server].hard,
-			   &servers[server].weight);
+			   &servers[server].weight, &servers[server].prio);
 		servers[server].sx = 0;
 		servers[server].rx = 0;
 		server++;
@@ -1272,9 +1282,10 @@ static void init(int argc, char **argv)
 	if (debuglevel) {
 		debug("servers:");
 		for (i = 0; i < nservers; i++) {
-			debug("%2d %s:%d:%d:%d", i,
+			debug("%2d %s:%d:%d:%d:%d:%d", i,
 				inet_ntoa(servers[i].addr), servers[i].port,
-				servers[i].maxc, servers[i].hard);
+				servers[i].maxc, servers[i].hard,
+				servers[i].weight, servers[i].prio);
 		}
 	}
 }
@@ -1385,7 +1396,7 @@ static void write_cfg(char *p)
 	if (foreground) fprintf(fp, " -f");
 	if (exit_enabled) fprintf(fp, " -X");
 	if (servers_max != SERVERS_MAX)
-		fprintf(fp, " -s %d", servers_max);
+		fprintf(fp, " -S %d", servers_max);
 	if (clients_max != CLIENTS_MAX)
 		fprintf(fp, " -c %d", clients_max);
 	if (e_server) fprintf(fp, " -e %s", e_server);
@@ -1427,10 +1438,13 @@ static void write_cfg(char *p)
 	else fprintf(fp, "no roundrobin\n");
 	for (i = 0; i < nservers; i++) {
 		fprintf(fp,
-			"server %d acl %d address %s port %d max %d hard %d\n",
+			"server %d acl %d address %s port %d max %d hard %d",
 			i, servers[i].acl,
 			inet_ntoa(servers[i].addr), servers[i].port,
 			servers[i].maxc, servers[i].hard);
+		if (weight) fprintf(fp, " weight %d", servers[i].weight);
+		if (prio) fprintf(fp, " prio %d", servers[i].prio);
+		fprintf(fp, "\n");
 	}
 	if (stubborn) fprintf(fp, "stubborn\n");
 	else fprintf(fp, "no stubborn\n");
@@ -1438,6 +1452,10 @@ static void write_cfg(char *p)
 	fprintf(fp, "tracking %d\n", tracking_time);
 	if (webfile) fprintf(fp, "web_stats %s\n", webfile);
 	else fprintf(fp, "no web_stats\n");
+	if (weight) fprintf(fp, "weight\n");
+	else fprintf(fp, "no weight\n");
+	if (prio) fprintf(fp, "prio\n");
+	else fprintf(fp, "no prio\n");
 	fclose(fp);
 }
 
@@ -1560,12 +1578,14 @@ static void do_cmd(char *b, void (*output)(char *, void *), void *op)
 			output(b, op);
 		}
 	} else if (!strcmp(p, "mode")) {
-		sprintf(b, "%sblock %sdelayed_forward %shash %sroundrobin %sstubborn\n",
+		sprintf(b, "%sblock %sdelayed_forward %shash %sroundrobin %sstubborn %sweight %sprio\n",
 			nblock?"no ":"",
 			delayed_forward?"":"no ",
 			hash?"":"no ",
 			roundrobin?"":"no ",
-			stubborn?"":"no ");
+			stubborn?"":"no ",
+			weight?"":"no ",
+			prio?"":"no ");
 		output(b, op);
 	} else if (!strcmp(p, "no")) {
 		p = strtok(NULL, " ");
@@ -1593,6 +1613,8 @@ static void do_cmd(char *b, void (*output)(char *, void *), void *op)
 			logfile = NULL;
 			if (logfp) fclose(logfp);
 			logfp = NULL;
+		} else if (!strcmp(p, "prio")) {
+			prio = 0;
 		} else if (!strcmp(p, "roundrobin")) {
 			roundrobin = 0;
 		} else if (!strcmp(p, "stubborn")) {
@@ -1605,6 +1627,8 @@ static void do_cmd(char *b, void (*output)(char *, void *), void *op)
 	} else if (!strcmp(p, "pid")) {
 		sprintf(b, "%ld\n", (long)getpid());
 		output(b, op);
+	} else if (!strcmp(p, "prio")) {
+		prio = 1;
 	} else if (!strcmp(p, "recent")) {
 		time_t when = time(NULL);
 		p = strtok(NULL, " ");
@@ -1640,14 +1664,16 @@ static void do_cmd(char *b, void (*output)(char *, void *), void *op)
 				servers[n].status = time(NULL)+atoi(q)-blacklist_time;
 			} else if (!strcmp(p, "weight")) {
 				servers[n].weight = atoi(q);
+			} else if (!strcmp(p, "prio")) {
+				servers[n].prio = atoi(q);
 			}
 		}
 	} else if (!strcmp(p, "servers")) {
 		for (n = 0; n < nservers; n++) {
-			sprintf(b, "%d addr %s port %d conn %d max %d hard %d weight %d sx %llu rx %llu\n",
+			sprintf(b, "%d addr %s port %d conn %d max %d hard %d weight %d prio %d sx %llu rx %llu\n",
 				n, inet_ntoa(servers[n].addr), servers[n].port,
 				servers[n].c, servers[n].maxc, servers[n].hard,
-				servers[n].weight,
+				servers[n].weight, servers[n].prio,
 				servers[n].sx, servers[n].rx);
 			output(b, op);
 		}
@@ -1758,7 +1784,7 @@ static int server_by_weight(void)
 		if (now-servers[i].status < blacklist_time || servers[i].weight == 0) {
 			continue;
 		}
-		load = servers[i].c/servers[i].weight;
+		load = (WEIGHT_FACTOR*servers[i].c)/servers[i].weight;
 		if (best_server == -1 || load < best_load) {
 			if (debuglevel)
 				debug("Server %d has load %d",
@@ -1771,6 +1797,30 @@ static int server_by_weight(void)
 	return best_server;
 }
 
+static int server_by_prio(void)
+{
+	int best_server = -1;
+	int best_prio = -1;
+	int i, prio;
+	int now = (int)time(NULL);
+
+	if (debuglevel) debug("server_by_prio()");
+	for (i = 0; i < nservers; i++) {
+		if (now-servers[i].status < blacklist_time) {
+			continue;
+		}
+		prio = servers[i].prio;
+		if (best_server == -1 || prio < best_prio) {
+			if (debuglevel)
+				debug("Server %d has prio %d", i, prio);
+			best_prio = prio;
+			best_server = i;
+		}
+	}
+	if (debuglevel) debug("Best prio server = %d", best_server);
+	return best_server;
+}
+
 static void add_client(int downfd, struct sockaddr_in *cli_addr)
 {
 	int upfd = -1, clino = -1, index = -1, n, pd;
@@ -1815,6 +1865,13 @@ static void add_client(int downfd, struct sockaddr_in *cli_addr)
 			}
 		}
 	}
+	if (prio) {
+		index = server_by_prio();
+		if (index != -1) {
+			upfd = try_server(index, 0, cli);
+			if (upfd != -1) goto Success;
+		}
+	}
 	if (weight) {
 		index = server_by_weight();
 		if (index != -1) {
@@ -1960,8 +2017,9 @@ static int flush_up(int i)
 
 static void mainloop_select(void)
 {
-	int downfd, clilen;
+	int downfd;
 	struct sockaddr_in cli_addr;
+	socklen_t clilen;
 	fd_set w_read, w_write, w_error;
 	int i, w_max;
 	usr1action.sa_handler = stats;
